#pragma once

//
//  member_ptr.hpp
//
//  This file is a derivative work from intrusive_ptr.hpp by Peter Dimov
//
//  Copyright (c) 2004 Oleg Fedtchenko
//
//  Permission to copy, use, modify, sell and distribute this software
//  is granted provided this copyright notice appears in all copies.
//  This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//

//
//  base classes used to work with member_ptr
//
//

//
//  base class for all members
//

class member_ptr_owner_base;

class member_ptr_member_base
{
    typedef member_ptr_member_base      this_type;
    typedef member_ptr_owner_base       owner_type;

    owner_type* pw_;

public:
    this_type( owner_type* pw) : pw_(pw){}
    virtual ~member_ptr_member_base(){}
    virtual owner_type* get_owner() const{ return pw_;}
};

//
//  base class for all owners
//

class member_ptr_owner_base : public member_ptr_member_base
{
    typedef member_ptr_member_base      base_type;
    typedef member_ptr_owner_base       this_type;

protected:
    int ref_count_;

public:

    this_type( this_type* pw) : base_type( (pw == (this_type*)0xFFFFFFFF) ? this : pw), ref_count_(0){}
    virtual ~member_ptr_owner_base(){}

    virtual void add_ref()
    {
        ++ref_count_;
    }

    virtual void release()
    {
        if (--ref_count_ == 0)
        {
            delete this;
        }
    }

    virtual int get_ref_count() const
    {
        return ref_count_;
    }
};

// 
// these functions are called by 
// intrusive_ptr<member_ptr_owner_base>
//

inline void intrusive_ptr_add_ref(member_ptr_owner_base * p)
{
    p->add_ref();
}
inline void intrusive_ptr_release(member_ptr_owner_base * p)
{
    p->release();
}

//
//  member_ptr
//
//  A smart pointer that uses intrusive-like reference counting for an owner of a class data-member.
//
//  The object is responsible for destroying itself
//  (if member_ptr_owner_base is not its direct ancestor).
//

template<class T> 
class member_ptr
{
private:

    typedef member_ptr this_type;

public:

    typedef T member_type;
    typedef member_ptr_owner_base owner_type;

    member_ptr(): px_(0), pw_(0)
    {
    }

    member_ptr(T * px, owner_type * pw, bool to_add_ref = true): px_(px), pw_(pw)
    {
        if( to_add_ref) add_ref();
    }

    member_ptr(T * px): px_(px), pw_(px->get_owner())
    {
        add_ref();
    }

    template<class U> member_ptr(member_ptr<U> const & rhs): px_(rhs.get()), pw_(rhs.get_owner())
    {
        add_ref();
    }

    member_ptr(member_ptr const & rhs): px_(rhs.px_), pw_(rhs.pw_)
    {
        add_ref();
    }

    ~member_ptr()
    {
        release_ref();
    }

    void add_ref()
    {
                           // to avoid exception when using member_ptr inside an owner ctor
        if(pw_ != 0  &&  ( ((void*)px_ == (void*)pw_)  ||  pw_->get_ref_count() > 0))
        {
            pw_->add_ref();
        }
    }

    void release_ref()
    {
        if(pw_ != 0  &&  pw_->get_ref_count() > 0)
        {
            pw_->release();
        }
    }

    template<class U> member_ptr & operator=(member_ptr<U> const & rhs)
    {
        this_type(rhs).swap(*this);
        return *this;
    }

    member_ptr & operator=(member_ptr const & rhs)
    {
        this_type(rhs).swap(*this);
        return *this;
    }

    T * get() const
    {
        return px_;
    }

    owner_type * get_owner() const
    {
        return pw_;
    }

    T & operator*() const
    {
        return *px_;
    }

    T * operator->() const
    {
        return px_;
    }

    typedef T * this_type::*unspecified_bool_type;

    operator unspecified_bool_type () const
    {
        return px_ == 0? 0: &this_type::px_;
    }

    // operator! is a Borland-specific workaround
    bool operator! () const
    {
        return px_ == 0;
    }

    void swap(member_ptr & rhs)
    {
        T * tmpX = px_;
        owner_type * tmpW = pw_;
        px_ = rhs.px_;
        pw_ = rhs.pw_;
        rhs.px_ = tmpX;
        rhs.pw_ = tmpW;
    }

private:

    T * px_;
    owner_type * pw_;
};

template<class T, class U> inline bool operator==(member_ptr<T> const & a, member_ptr<U> const & b)
{
    return a.get() == b.get();
}

template<class T, class U> inline bool operator!=(member_ptr<T> const & a, member_ptr<U> const & b)
{
    return a.get() != b.get();
}

template<class T> inline bool operator==(member_ptr<T> const & a, T * b)
{
    return a.get() == b;
}

template<class T> inline bool operator!=(member_ptr<T> const & a, T * b)
{
    return a.get() != b;
}

template<class T> inline bool operator==(T * a, member_ptr<T> const & b)
{
    return a == b.get();
}

template<class T> inline bool operator!=(T * a, member_ptr<T> const & b)
{
    return a != b.get();
}

#if __GNUC__ == 2 && __GNUC_MINOR__ <= 96

// Resolve the ambiguity between our op!= and the one in rel_ops

template<class T> inline bool operator!=(member_ptr<T> const & a, member_ptr<T> const & b)
{
    return a.get() != b.get();
}

#endif

template<class T> inline bool operator<(member_ptr<T> const & a, member_ptr<T> const & b)
{
    return std::less<T *>()(a.get(), b.get());
}

template<class T> void swap(member_ptr<T> & lhs, member_ptr<T> & rhs)
{
    lhs.swap(rhs);
}

// mem_fn support

template<class T> T * get_pointer(member_ptr<T> const & p)
{
    return p.get();
}
template<class T, class U> member_ptr<T> static_pointer_cast(member_ptr<U> const & p)
{
    return member_ptr<T,W>( static_cast<T *>(p.get()), static_cast<member_ptr<T>::owner_type *>(p.get_owner()));
}

template<class T, class U> member_ptr<T> const_pointer_cast(member_ptr<U> const & p)
{
    return member_ptr<T,W>( const_cast<T *>(p.get()), const_cast<member_ptr<T>::owner_type *>(p.get_owner()));
}

template<class T, class U> member_ptr<T> dynamic_pointer_cast(member_ptr<U> const & p)
{
    return member_ptr<T>( dynamic_cast<T *>(p.get()), dynamic_cast<member_ptr<T>::owner_type *>(p.get_owner()));
}

#ifdef _MSVC
# pragma warning(pop)
#endif
